#include "Rock.h"
#include "Stage.h"
#include "Stages.h"
#include "Vram.h"
#include "Chars.h"
#include "Sprite.h"
#include "Direction.h"
#include "Man.h"
#include "Point.h"
#include "Chaser.h"
#include "Ghost.h"
#include "Sound.h"

extern void _deb();
constexpr byte RateShift = 4;
constexpr byte MaxRate = 3;
constexpr byte RateMask = 0x30;

Movable[MaxRockCount] Rocks;

static ptr<byte> MapP(ptr<Movable> pRock)
{
    return TerrainMapPtr(pRock->x >> MapShift, pRock->y >> MapShift);
}


ptr<byte> InitRocks(ptr<byte> pStage)
{
    ptr<Movable> pRock;
    for (pRock: Rocks) {
        pRock->dy = 0;
    }
    return InitMovables(Rocks, Sprite_Rock, MaxRockCount, pStage);
}


void DrawRocks()
{
    ptr<Movable> pRock;
    for (pRock: Rocks) {
        if ((pRock->status & Movable_Status_Live) == 0) continue;
        ShowSprite(pRock, Pattern_Rock);
        *MapP(pRock) |= Terrain_Rock;
    }
}

// static void BeginFalling(ptr<Movable> pRock)
// {
//     // ptr<Movable> pRock;
//     // for (pRock: Rocks) {
//     //     if ((pRock->status & Movable_Status_Live) != 0) continue;
//     //     pRock->x = x << CoordShift;
//     //     pRock->y = y << CoordShift;
//         pRock->status = Rock_Status_Falling | Movable_Status_CanMove;
//         DrawFalling(pRock);
//         // return true;
//     // }
//     // return false;
// }


void TestRocks()
{
    constexpr byte Size = 2;
    ptr<Movable> pRock;
    for (pRock: Rocks) {
        byte below, x, y;
        if (pRock->dy != 0) continue;
        if ((pRock->status & Movable_Status_Live) == 0) continue;
        below = CurrentTerrain(pRock->x >> MapShift, (pRock->y >> MapShift) + 1);
        if ((below & Terrain_Mask) != Terrain_Space) continue;
        if (
            (man.y >> CoordShift) - Size == (pRock->y >> CoordShift) &&
            (man.x >> CoordShift) - (pRock->x >> CoordShift) + Size / 2 < Size
        ) continue;
        pRock->status = Movable_Status_Live | Movable_Status_CanMove;
        pRock->dy = 1;
        *MapP(pRock) &= ~Terrain_Rock;
        SetDirection(pRock, Direction_Down);
    }
}


static bool IsNearAny(ptr<Movable> pRock, byte x, byte y)
{
    if (IsNearXY(&man, x, y)) return true;
    if (IsNearRock(pRock, x, y)) return true;
    return false;
}


void MoveRocks()
{
    constexpr byte Size = 2 << CoordShift;
    ptr<Movable> pRock;
    for (pRock: Rocks) {
        if (pRock->dy == 0) continue;
        if (CanChangeDirection(pRock)) {
            byte x, y;
            ptr<byte> pCurrent, pBelow;
            byte current, below;
            x = pRock->x >> MapShift;
            y = pRock->y >> MapShift;
            if (y >= MapHeight - 1) goto end;
            pBelow = TerrainMapPtr(x, y + 1);
            below = *pBelow;
            if ((below & Terrain_Mask) != Terrain_Space) {
                end:
                pRock->dy = 0;
                *MapP(pRock) |= Terrain_Rock;
                continue;
            }
            if (IsNearAny(pRock, pRock->x, pRock->y + Size)) {
                continue;
            }
            pCurrent = TerrainMapPtr(x, y);
            current = *pCurrent & ~Wall_Bottom;
            *pCurrent = current;
            DrawTerrain(x, y, current);
            below &= ~Wall_Top;
            *pBelow = below;
            DrawTerrain(x, y + 1, below);
        }
        if (IsOnGrid(pRock)) {
            HitChaser(pRock);
            HitGhost(pRock);
        }
        MoveMovable(pRock);
        ShowSprite(pRock, Pattern_Rock);
    }
}


bool IsNearRock(ptr<Movable> pMovable, byte x, byte y)
{
    ptr<Movable> pRock;
    for (pRock: Rocks) {
        if ((pRock->status & Movable_Status_Live) == 0) continue;
        if (pRock != pMovable && IsNearXY(pRock, x, y)) return true;
    }
    return false;
}


void OnHitRock(ptr<Movable> pRock, ptr<Movable> pMonster)
{
    byte rate;
    rate = pRock->status >> RateShift;
    if (rate < MaxRate + 1) {
        ++rate;
        pRock->status =
                (pRock->status & ~RateMask) | (rate << RateShift);
    }
    StartPoint(pMonster->x, pMonster->y, rate);
    // UpdateSprites();
    Sound_Hit();
}


bool AdjacentRock(ptr<Movable> pMovable, byte x, byte y)
{
    AdjacentOtherInArray(Rocks, MaxRockCount, pMovable, x, y);
}